#! /usr/bin/env python3

__author__  = 'Felis-Sapiens'

import argparse
import struct
import zlib
from collections import namedtuple
from hashlib import sha256
from io import BytesIO
from re import sub
from Crypto.Cipher import AES

ENDIANNESS = '>'
SIGN_HDR_FMT = ENDIANNESS + '3I'
ENCRY_HDR_FMT = ENDIANNESS + '7I32s'
CHUNK_HDR_FMT = ENDIANNESS + '3I'

SignHdr = namedtuple('SignHdr', 'magic unk04 size')
EncryHdr = namedtuple('EncryHdr', 'magic type unk08 unk0C buf_size data_crc self_crc unk1C')
ChunkHdr = namedtuple('ChunkHdr', 'out_size in_size not_last')

KEYS = [
    b'MIK@0STzKpB%qJZe',
    b'MIK@0STzKpB%qJZf',
    b'402c38de39bed665',
    b'Wj\0\0\0\0\0\0\0\0\0\0\0\0\0\0',
    b'm8@96&ZG3Nm7N&Iz',
    b'GrWM2Hz&LTvz&f^5',
    b'GrWM3Hz&LTvz&f^9',
    b'Renjx%2$CjM\0\0\0\0\0',
    b'tHG@Ti&GVh@ql3XN',
    b'SDEwOE5WMi41Uk9T']

def read_sign_header(fd):
    old_pos = fd.tell()
    hdr_size = struct.calcsize(SIGN_HDR_FMT)
    bs = fd.read(hdr_size)
    if len(bs) != hdr_size:
        print('Failed to read SignHdr: insufficient data')
        fd.seek(old_pos)
        return
    sign_hdr = SignHdr._make(struct.unpack(SIGN_HDR_FMT, bs))
    if sign_hdr.magic != 0x4030201:
        print('Wrong SignHdr.magic: {:08X} != {:08X}. The file is probably not signed.'.format(sign_hdr.magic, 0x4030201))
        fd.seek(old_pos)
        return
    return sign_hdr

def read_encry_header(fd):
    hdr_size = struct.calcsize(ENCRY_HDR_FMT)
    bs = fd.read(hdr_size)
    if len(bs) != hdr_size:
        print('ERROR: failed to read EncryHdr (insufficient data)')
        return
    encry_hdr = EncryHdr._make(struct.unpack(ENCRY_HDR_FMT, bs))
    print('encryption header magic:       0x{:08X}'.format(encry_hdr.magic))
    if encry_hdr.magic != 0x01020304:
        print('ERROR: expected magic is 0x01020304')
        return
    print('encryption type:              ', encry_hdr.type)
    if encry_hdr.self_crc != 0:
        print('encryption header CRC32:       0x{:08X}'.format(encry_hdr.self_crc))
        self_crc = zlib.crc32(bs[:0x18])
        if encry_hdr.self_crc != self_crc:
            print("ERROR: CRC32 doesn't match")
            return
    return encry_hdr

def read_chunk_header(fd):
    hdr_size = struct.calcsize(CHUNK_HDR_FMT)
    bs = fd.read(hdr_size)
    if len(bs) != hdr_size:
        print('ERROR: failed to read chunk header')
        return
    chunk_hdr = ChunkHdr._make(struct.unpack(CHUNK_HDR_FMT, bs))
    print('chunk input length:           ', hex(chunk_hdr.in_size))
    print('chunk output length:          ', hex(chunk_hdr.out_size))
    return chunk_hdr

def read_compressed_data(fd, encry_hdr):
    crc = 0
    fd_out = BytesIO()
    while True:
        chunk_hdr = read_chunk_header(fd)
        if chunk_hdr is None:
            return
        chunk = fd.read(chunk_hdr.in_size)
        if len(chunk) != chunk_hdr.in_size:
            print('ERROR: failed to read compression chunk data')
            return
        crc = zlib.crc32(chunk, crc)
        chunk = zlib.decompress(chunk)
        if len(chunk) != chunk_hdr.out_size:
            print('ERROR: wrong length of uncompressed data')
            return
        fd_out.write(chunk)
        if chunk_hdr.not_last == 0:
            break
    
    print('compressed data CRC32:         0x{:08X}'.format(encry_hdr.data_crc))
    if encry_hdr.data_crc != 0 and encry_hdr.data_crc != crc:
        print("ERROR: CRC32 doesn't match")
        return
    fd_out.seek(0)
    return fd_out

def read_aes_data(fd, key):
    enc_data = b''
    while True:
        chunk_hdr = read_chunk_header(fd)
        if chunk_hdr is None:
            return
        chunk = fd.read(chunk_hdr.in_size)
        if len(chunk) != chunk_hdr.in_size:
            print('ERROR: failed to read AES chunk data')
            return
        enc_data += chunk
        if chunk_hdr.not_last == 0:
            break
    
    keys = KEYS[:]
    if key:
        keys = [key.ljust(16, b'\0')[:16]] + KEYS
    else:
        keys = KEYS[:]
    key = None
    for key_ in KEYS:
        aes_cipher = AES.new(key_, AES.MODE_ECB)
        if aes_cipher.decrypt(enc_data[:16]).startswith(b'\x01\x02\x03\x04'):
            print('key:                          ', key_.decode('Latin1', errors='backslashreplace'))
            key = key_
            break
    if key is None:
        print('ERROR: failed to decrypt (expected magic is 0x01020304), please specify the correct key with --key')
        return
    aes_cipher = AES.new(key, AES.MODE_ECB)
    fd_out = BytesIO()
    fd_out.write(aes_cipher.decrypt(enc_data))
    fd_out.seek(0)
    return fd_out

def derive_aes_key(key_seed, iv_seed):
    return sha256(key_seed).digest()[:32], sha256(iv_seed).digest()[:16]

def read_aes_data2(fd, model_name, SN, MAC, encry_key):
    enc_data = b''
    key = None
    iv = None
    fd_out = BytesIO()
    mac_str = ':'.join(f'{b:02x}' for b in MAC)
    vendor_specific = SN[4:]
    while True:
        chunk_hdr = read_chunk_header(fd)
        if chunk_hdr is None:
            return
        chunk = fd.read(chunk_hdr.in_size)
        if len(chunk) != chunk_hdr.in_size:
            print('ERROR: failed to read AES chunk data')
            return
        if key is None:
            key_, iv_ = derive_aes_key(model_name + b'Key02721401', model_name + b'Iv02721401')
            aes_cipher = AES.new(key_, iv=iv_, mode=AES.MODE_CBC)
            if aes_cipher.decrypt(chunk[:16]).startswith(b'\x01\x02\x03\x04'):
                print('decrypted with ModelName:     ', model_name.decode('Latin1', errors='backslashreplace'))
                key = key_
                iv = iv_
        if key is None:
            key_, iv_ = derive_aes_key(encry_key + SN + b'Mcd5c46e', b'G21b667b' + mac_str.encode('Latin1') + encry_key)
            aes_cipher = AES.new(key_, iv=iv_, mode=AES.MODE_CBC)
            if aes_cipher.decrypt(chunk[:16]).startswith(b'\x01\x02\x03\x04'):
                print('decrypted with SN, MAC, key:  ', SN.decode('Latin1'), mac_str, encry_key.decode('Latin1'))
                key = key_
                iv = iv_
        if key is None:
            key_, iv_ = derive_aes_key(b'MTKen7513key', b'MTKen7513IV')
            aes_cipher = AES.new(key_, iv=iv_, mode=AES.MODE_CBC)
            if aes_cipher.decrypt(chunk[:16]).startswith(b'\x01\x02\x03\x04'):
                key = key_
                iv = iv_
        if key is None:
            key_, iv_ = derive_aes_key(b'MTKen7513key@Sor*3', b'MTKen7513IV@Sor*3#a')
            aes_cipher = AES.new(key_, iv=iv_, mode=AES.MODE_CBC)
            if aes_cipher.decrypt(chunk[:16]).startswith(b'\x01\x02\x03\x04'):
                key = key_
                iv = iv_
        if key is None:
            key_, iv_ = derive_aes_key(b'rtYo0PV@Sor*3#a6&ZG3Nm7N&I', b'1liy_t@6&ZG3Nm7N&IV@Sor*3#a')
            aes_cipher = AES.new(key_, iv=iv_, mode=AES.MODE_CBC)
            if aes_cipher.decrypt(chunk[:16]).startswith(b'\x01\x02\x03\x04'):
                key = key_
                iv = iv_
        if key is None:
            key_, iv_ = derive_aes_key((b'0PV@Sor*3#a6&ZG3Nm7N' + SN)[:31], (b't@6&ZG7N&IV@Sor*3' + SN)[:31])
            aes_cipher = AES.new(key_, iv=iv_, mode=AES.MODE_CBC)
            if aes_cipher.decrypt(chunk[:16]).startswith(b'\x01\x02\x03\x04'):
                key = key_
                iv = iv_
        if key is None:
            key_, iv_ = derive_aes_key(b'L04&Product@5A238dc79b15726d5c06', b'ZTE%FN$GponNJ025678b02a85c63c706')
            aes_cipher = AES.new(key_, iv=iv_, mode=AES.MODE_CBC)
            if aes_cipher.decrypt(chunk[:16]).startswith(b'\x01\x02\x03\x04'):
                key = key_
                iv = iv_
        if key is None:
            key_, iv_ = derive_aes_key(vendor_specific + MAC[::-1].hex().encode('Latin1'), b'ZTE%FN$GponNJ025678b02a85c63c706')
            aes_cipher = AES.new(key_, iv=iv_, mode=AES.MODE_CBC)
            if aes_cipher.decrypt(chunk[:16]).startswith(b'\x01\x02\x03\x04'):
                print('decrypted with SN, MAC:  ', SN.decode('Latin1'), mac_str)
                key = key_
                iv = iv_
        if key is None:
            key_, iv_ = derive_aes_key(b'f688v9.0', b'ZTE%FN$GponNJ025')
            aes_cipher = AES.new(key_, iv=iv_, mode=AES.MODE_CBC)
            if aes_cipher.decrypt(chunk[:16]).startswith(b'\x01\x02\x03\x04'):
                key = key_
                iv = iv_
        if key is None:
            key_, iv_ = derive_aes_key(vendor_specific + MAC[::-1].hex().encode('Latin1'), b'ZTE%FN$GponNJ025')
            aes_cipher = AES.new(key_, iv=iv_, mode=AES.MODE_CBC)
            if aes_cipher.decrypt(chunk[:16]).startswith(b'\x01\x02\x03\x04'):
                print('decrypted with SN, MAC:  ', SN.decode('Latin1'), mac_str)
                key = key_
                iv = iv_
        if key is None:
            key_, iv_ = derive_aes_key(b'f2801sv2', b'ZTE%FN$GponNJ025')
            aes_cipher = AES.new(key_, iv=iv_, mode=AES.MODE_CBC)
            if aes_cipher.decrypt(chunk[:16]).startswith(b'\x01\x02\x03\x04'):
                key = key_
                iv = iv_
        if key is None:
            print('ERROR: failed to decrypt (expected magic is 0x01020304), please specify the correct --model or --SN, --MAC, --encry_key')
            return
        aes_cipher = AES.new(key, iv=iv, mode=AES.MODE_CBC)
        fd_out.write(aes_cipher.decrypt(chunk))
        if chunk_hdr.not_last == 0:
            break
    
    fd_out.seek(0)
    return fd_out

def read_config(args):
    fd_in = args.infile
    fd_out = args.outfile
    if fd_in.read(16) == b'\x99\x99\x99\x99\x44\x44\x44\x44\x55\x55\x55\x55\xAA\xAA\xAA\xAA':
        fd_in.seek(0x80)
    else:
        fd_in.seek(0)
    sign_hdr = read_sign_header(fd_in)
    if sign_hdr is not None:
        sign_key = fd_in.read(sign_hdr.size)
        if len(sign_key) != sign_hdr.size:
            print('Failed to read SignKey: insufficient data')
            return
        print('sign key:                     ', sign_key.decode('Latin1', errors='backslashreplace'))
    encry_hdr = read_encry_header(fd_in)
    enc_type = encry_hdr.type
    
    if enc_type == 3 or enc_type == 4 or enc_type == 5 or enc_type == 6:
        fd_in = read_aes_data2(fd_in, args.model, args.SN, args.MAC, args.encry_key)
        if fd_in is None:
            return
    
    if enc_type == 2 or enc_type == 1:
        fd_in = read_aes_data(fd_in, args.key)
        if fd_in is None:
            return
    
    if enc_type >= 2:
        encry_hdr = read_encry_header(fd_in)
        enc_type = 0
    
    if enc_type == 0:
        fd_in = read_compressed_data(fd_in, encry_hdr)
        if fd_in is None:
            return
    
    fd_out.write(fd_in.read())
    return

def str_to_mac(s):
    mac = sub('[^0-9A-Fa-f]', '', s)
    if len(mac) != 12:
        raise argparse.ArgumentTypeError(f'Invalid MAC: {s}')
    return bytes.fromhex(mac)

def main():
    parser = argparse.ArgumentParser(description='Decode configuration file '\
                                     '(config.bin) from ZTE routers',
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     epilog="""\
Known AES keys:
  zxhn h118n ert5                      - 'MIK@0STzKpB%qJZe'
  zxhn h118n V2.1.3_ROSCNT?            - 'MIK@0STzKpB%qJZf'
  zxhn h168n v3                        - '402c38de39bed665'
  zxhn h298n hv17_fv116_mts?t1         - 'Wj' (due to bug, orig. is 'Wj%2$CjM')
  zxhn h298a hw1.1.20_fw1.1.20_ros_t1? - 'm8@96&ZG3Nm7N&Iz'
  ZXHN H108N 2.5.5_RU8T9               - 'SDEwOE5WMi41Uk9T'
  zxhn h108n hw1.2_fw2.5.4_eg1t8_ted,
  zxhn h108n hv11_fv2_5_4_*            - 'GrWM2Hz&LTvz&f^5'
  zxhn h168n hv10_fv310t3_belt         - 'GrWM3Hz&LTvz&f^9'
  zxhn h208n hv10_fv1010_belt16t1      - 'Renjx%2$CjM'
  zxhn h267n hv10_fv100t3_belt         - 'tHG@Ti&GVh@ql3XN'

Examples:
  """+__file__+""" config.bin config.xml --key 402c38de39bed665
  """+__file__+""" config.bin config.xml --SN ZTERRUVK8Q05390 --MAC c0:fd:84:d4:d5:7f
  """+__file__+""" config.bin config.xml --model ZXHNH168NV3.5""")
    
    parser.add_argument('infile', type=argparse.FileType('rb'),
                        help='Encoded configuration file (config.bin)')
    parser.add_argument('outfile', type=argparse.FileType('wb'),
                        help='Output file')
    parser.add_argument('--key', type=lambda x:x.encode(), default=b'',
                        help="Key for AES encryption")
    parser.add_argument('--model', type=lambda x:x.encode(), default=b'',
                        help="ModelName without spaces")
    parser.add_argument('--SN', type=lambda x:x.encode(), default=b'',
                        help="SerialNumber")
    parser.add_argument('--MAC', type=str_to_mac, default=b'',
                        help="MAC address")
    parser.add_argument('--encry_key', type=lambda x:x.encode(), default=b'',
                        help="EncryKey")
    args = parser.parse_args()
    
    read_config(args)

if __name__ == '__main__':
    main()